/**
 * @version 3.5 2004-04-12
 * @author Gonzalo Ojeda 9821018-0
 */
 
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;

/**
 * La clase <b>FigPanel</b> corresponde a la zona de dibujo, donde se ubican 
 * las figuras y se realizan las operaciones sobre estas mismas. 
 */
public class FigPanel extends JPanel
{
   /**
    * Constructor de la clase <b>FigPanel</b>, define los ActionListener para 
    * los botones de Figuras y de Operaciones, ademas realiza todas las 
    * operaciones, tanto la creacion de las figuras como las operaciones sobre
    * ellas.
    * @param rectButton Boton que indica que la proxima accion es dibujar 
    * <i>Rectangulos</i>
    * @param elipButton Boton que indica que la proxima accion es dibujar 
    * <i>Elpses</i>
    * @param linButton Boton que indica que la proxima accion es dibujar 
    * <i>Lneas</i>
    * @param rotButton Boton que indica que la proxima accion es <i>Rotar</i> 
    * el dibujo
    * @param escButton Boton que indica que la proxima accion es <i>Escalar</i> 
    * el dibujo
    * @param trasButton Boton que indica que la proxima accion es 
    * <i>Trasladar</i> el dibujo
    */
    public FigPanel (JButton rectButton, JButton elipButton, JButton linButton,
    JButton rotButton, JButton escButton, JButton trasButton)
    {
      JPanel fig = new JPanel();
      // arreglo de largo variable para almacenar todas las figuras a crear
      shapes = new ArrayList();
      // la figura actual, se la inicializa vaca.
      current = null;
      // necesaria para las operaciones que funcionan con mas de un click
      contador = 0;
      // punto cualquiera, la idea es tener un valor para comparar y
      // comprobar que se han hecho 2 clicks del mouse y no 1
      nullPoint = new Point2D.Double (0, 0);
      // se crean listener para todas las operaciones del mouse.
      addMouseListener(new MouseHandler());
      addMouseMotionListener (new MouseMotionHandler());
      // ActionListener para responder a los botones de Figura y Operacion,
      // almacena la accion realizada en un string llamado figura, y limpia el
      // almacena
      ActionListener figListener = new ActionListener ()
      {
        public void actionPerformed (ActionEvent e)
        {
          figura = e.getActionCommand();
          contador = 0;
        }
      };
      // se agrega el mismo action listener a los 6 botones, pues las 6 
      // acciones son excluyentes entre s.
      rectButton.addActionListener(figListener);
      elipButton.addActionListener(figListener);
      linButton.addActionListener(figListener);
      rotButton.addActionListener(figListener);
      escButton.addActionListener(figListener);
      trasButton.addActionListener(figListener);
    }
   /**
    * Funcion que es invocada por la maquina virtual cuando se realiza una 
    * peticion de repintado (<i>repaint()</i>), invoca a la funcion 
    * paintComponent de la clase superior. Esta misma funcion realiza la 
    * rotacion, escalamiento, o traslacion del dibujo completo en caso de ser
    * solicitado (click en el respectivo boton) sin embargo, estos 3 cambios 
    * son solo temporales hasta que no se inserte una nueva figura, pues los
    * cambios se hacen sobre el objeto grafico <i>g2</i> y no sobre el arreglo
    * de las figuras, por la imposibilidad de guardar figuras rotadas.
    * @param g   Objeto entregado por la maquina virtual que permite dibujar las
    *            figuras previo casteo a un objeto de tipo Graphics2D
    */
    public void paintComponent(Graphics g)
    {
      super.paintComponent(g);
      Graphics2D g2 = (Graphics2D)g;
      // solo si esta marcado el boton Rotar, se realiza esta accion. 
      if ("Rotar".equals(figura))
        g2.rotate(tetha, iniRot.getX(), iniRot.getY());
      // solo si esta marcado el boton Escalar, se realiza esta accion.
      else if ("Escalar".equals(figura))
        g2.scale(zoom, zoom);
      // solo si esta marcado el boton Trasladar, se realiza esta accion.
      else if ("Trasladar".equals(figura))
        g2.translate(dx, dy);
      // dibujar todas las formas contenidas en el arreglo de largo variable
      for (int i = 0; i < shapes.size(); i++)
        g2.draw((Shape)shapes.get(i));
    }

   /**
    * Funcion que agrega una nueva figura al arreglo de largo variable, segn 
    * sea la opcion seleccionada anteriormente.
    * La funcion trabaja sobre 2 puntos, al realizar el primer click se definen
    * estos 2 puntos al mismo tiempo, y a medida que se mueve el mouse con el
    * boton presionado se va variando el segundo punto.
    * Las formas creadas son <i>Rectangle2D.Double</i> para Rectangulos,
    * <i>Ellipse2D.Double</i> para las Elpses, y <i>Line2D.Double</i> para las 
    * lneas, estas 3 clases heredan de la Interfase <i>Shape</i>, por lo que 
    * se puede usar un objeto de este tipo para trabajar con los 3 tipos de 
    * figura.
    */
    public void add ()
    {
      double a = inicio.getX();
      double b = inicio.getY();
      double x = last.getX();
      double y = last.getY();
      // Crear Rectangulos, luego de crearlo y agregarlo al arreglo, se 
      // redibuja todo.
      if ("Rectangulo".equals(figura))
      {
        current = new Rectangle2D.Double (a, b, x - a, y - b);
        shapes.add(current);
        repaint();
      }
      // Crear Elpses, luego de crearlo y agregarlo al arreglo, redibujar todo
      else if ("Elipse".equals(figura))
      {
        current = new Ellipse2D.Double(a, b, x - a, y - a);
        shapes.add(current);
        repaint();
      }
      // Crear Lneas, luego de crearlo y agregarlo al arreglo, redibujar todo
      else if ("Linea".equals(figura))
      {
        current = new Line2D.Double(inicio, last);
        shapes.add(current);
        repaint();
      }
    }
   /**
    * Funcion que limpia la zona de dibujo, borra todos los elementos del 
    * arreglo de figuras y actualiza la zona de dibujo con ningn elemento.
    * Este mtodo es invocado al presionar el boton limpiar, y es llamado desde
    * el otro archivo. (por la clase <i>EditorFrame</i>), tambien limpia el 
    * contador.
    */
    public void clearShapes()
    {
      shapes.clear();
      contador = 0;
      repaint();
    }

   /** 
    * Clase interna que maneja los clicks del mouse, extiende la clase 
    * MouseAdapter por lo tanto solo es necesario definir las funciones que se
    * van a utilizar del mouse, en este caso solo mousePressed.
    */
    private class MouseHandler extends MouseAdapter
    {
     /**
      * Mtodo llamado al presionar el mouse, al presionar el mouse se agregan 
      * el punto inicial y final de la figura (todas las formas a dibujar 
      * requieren de una u otra forma 2 puntos para ser dibujadas).
      * En el caso de tratarse de una figura, se definen los 2 puntos altiro, y
      * se llama a la funcion <i>add()</i> para crear la figura.
      * Para el caso de Rotar se define un punto para el primer click, y otro
      * para el segundo para despues en base a el movimiento del mouse calcular
      * el angulo de rotacion. Para el caso de Escalar se define un punto al 
      * primer click y el segundo punto al siguiente click, y segun el 
      * movimiento del mouse se hace "zoom".
      */
      public void mousePressed (MouseEvent event)
      {
        if (("Rectangulo".equals(figura)) || ("Elipse".equals(figura)) 
        || ("Linea".equals(figura)))
        {
          inicio = event.getPoint();
          last = event.getPoint();
          add();
        }
        else if ("Rotar".equals(figura))
        {
          if (contador == 0)
          {
            iniRot = event.getPoint();
            // al hacer el primer click se hace esto porque hay problemas se
            // intenta rotar al presionar el boton por primer vez y arrastrar
            rotPoint = nullPoint;
            contador = 1;
          }
          else
          {
            rotPoint = event.getPoint();
            contador--;
          }
        }
        else if ("Escalar".equals(figura))
        {
          if (contador == 0) 
          {
            trasPoint = event.getPoint();
            contador = 1;
            // para impedir que solo se presione una vez el mouse
            // para escalar (se necesitan 2 clicks
            iniDistancia = 0;
          }
          else
          {
            last = event.getPoint();
            contador--;
            // Se calcula la distancia inicial si ya se tienen los 2 puntos
            iniDistancia = trasPoint.distance(last);
          }
        }
        else if ("Trasladar".equals(figura))
        {
          trasPoint = event.getPoint();
        }
      }
      public void mouseReleased(MouseEvent event)
      {
        if ("Trasladar".equals(figura))
        {
          dx1 = dx;
          dy1 = dy;
        }
      }
    }
   /**
    * Clase Interna para manejar los movimientos del mouse, en este caso el 
    * mtodo <i>mouseMoved</i> que corresponde a movimientos del mouse sin 
    * presionar el botonno es utilizado, as que no se implementa.
    * El mtodo <i>mouseDragged</i> se utiliza para actualizar las figuras, y 
    * realizar las operaciones mientras se mueve el mouse.
    * Lo primero que se hace es obtener el punto donde esta actualmente el 
    * mouse y guardarlo.
    * Para la creacion de Rectangulos y Elpses se realiza un 
    * setFrameFromDiagonal, para actualizar la figura a medida que se mueve el 
    * mouse. Para lneas se usa setLine. Para las operaciones se obtiene el 
    * punto final, se realizan algunos calculos en caso de ser necesarios, y se
    *  llama al mtodo <i>repaint()</i> para actualizar el dibujo.
    */
    private class MouseMotionHandler implements MouseMotionListener
    {
      public void mouseMoved (MouseEvent event) 
      {
      };
      public void mouseDragged (MouseEvent event)
      {
        last = event.getPoint();
        if ("Rectangulo".equals(figura))
        {
          Rectangle2D rect = (Rectangle2D)current;
          rect.setFrameFromDiagonal(inicio, last);
          repaint();
        }
        else if ("Elipse".equals(figura))
        {
          Ellipse2D elip = (Ellipse2D)current;
          elip.setFrameFromDiagonal(inicio, last);
          repaint();
        }
        else if ("Linea".equals(figura))
        {
          Line2D lin = (Line2D)current;
          lin.setLine (inicio, last);
          repaint();
        }
        else if ("Trasladar".equals(figura))
        {
          dx = event.getX() - trasPoint.getX();
          dy = event.getY() - trasPoint.getY();
          dx = dx1 + dx;
          dy = dy1 + dy;
          repaint();
        }
        else if ("Escalar".equals(figura))
        {
          // no escala si solo se ha presionado una vez el mouse
          if (iniDistancia == 0)
            return;
          // se calcula la distancia desde el punto inicial al punto actual
          finDistancia = trasPoint.distance(last);
          // calcular el factor para hacer zoom dividiendo la distancia actual
          // por la distancia inicial entre los 2 puntos.
          zoom = finDistancia / iniDistancia;
          repaint();
        }
        else if ("Rotar".equals(figura))
        {
          // si solo ha habido un click, no rotar.
          if (rotPoint.equals(nullPoint))
            return;
          // el calculo del angulo se realiza por medio del teorema del coseno
          // la variable temp se uso para calcular b^2 + c^2 - a^2
          double temp = iniRot.distanceSq(last) + iniRot.distanceSq(rotPoint) 
          - rotPoint.distanceSq(last);
          tetha = Math.acos(temp / (2 * iniRot.distance(last) 
          * iniRot.distance(rotPoint)));
          // El calculo anterior solo devuelve valores positivos, para obtener
          // angulos negativos cuando corresponda se realiza lo siguiente.
          if (last.getY() < rotPoint.getY())
            tetha = -tetha;
          repaint();
        }
      }  
    }

  private ArrayList shapes;  // Arreglo de largo variable almacena las figuras
  private Shape current;     // figura actual
  // string que almacena que se esta realizando segn botones
  private String figura;
  private int contador;
  private double tetha;   // angulo para rotaciones.
  // variables para las operaciones.
  private double iniDistancia, finDistancia, zoom;
  private Point2D inicio;  // punto inicial de las figuras
  private Point2D  last;   // punto final de las figuras
  private Point2D trasPoint, rotPoint, iniRot;  // puntos para las operaciones
  private Point2D nullPoint;  // punto para que no haya errores en rotar
  private double dx, dy, dx1 , dy1;  // deltas para el trasladar
  private Point2D P1;
  }
